@@ -4,16 +4,16 @@ from datetime import datetime, timedelta
 from time import sleep
 from typing import Union
 
+import certifi
 import pandas as pd
 import requests
 from dateutil.tz import tzutc
-from powerbot_client import Configuration, ApiClient, ContractApi, Signal, Trade, \
-    InternalTrade, OwnOrder
+from powerbot_client import Configuration, ApiClient, ContractApi, Signal, Trade, InternalTrade, OwnOrder, ContractItem
+from tqdm import tqdm
 
 from powerbot_backtesting.historic_data_acquisition import get_historic_contract_ids
 from powerbot_backtesting.models import HistoryApiClient
-from powerbot_backtesting.utils import _cache_data, _check_contracts, _get_file_cachepath, _find_cache, \
-    _get_private_data
+from powerbot_backtesting.utils import _cache_data, _check_contracts, _get_file_cachepath, _find_cache, _get_private_data
 from powerbot_backtesting.utils.constants import *
 
 
@@ -28,7 +28,7 @@ def init_client(api_key: str, host: str) -> ApiClient:
     Returns:
         PowerBot ApiClient Object
     """
-    config = Configuration(api_key={'api_key_security': api_key}, host=host)
+    config = Configuration(api_key={'api_key_security': api_key}, host=host, ssl_ca_cert=certifi.where())
     return ApiClient(config)
 
 
@@ -39,21 +39,11 @@ def get_contracts(api_client: Union[ApiClient, HistoryApiClient],
                   contract_time: str = "all",
                   products: list[str] = None,
                   allow_udc: bool = False,
-                  delivery_areas: list[str] = None):
+                  delivery_areas: list[str] = None) -> list[ContractItem]:
     """
-    Loads all contract IDs for specified timeframe, either by hours, single day or multiple days. Alternatively, a list
-    of contract IDs can be passed instead of a timeframe, returning a dictionary of contract IDs compatible with all other
-    functions of the Backtesting package. If return_contract_objects is True, a dictionary of contract objects will be
-    returned as well.
+    Loads all contracts for specified timeframe. Alternatively, a list of contract IDs can be passed instead of a timeframe.
 
-    If a historic API client is used all parameters except time_from, time_till, contract_time, products and
-    allow_udc will be ignored.
-
-    If time_from includes hh:mm:ss
-        loads data between both times in timesteps according to contract_time
-
-    If time_from doesn't include hh:mm:ss
-        loads data between both dates in timesteps according to contract_time
+    If a historic API client is used all parameters except time_from, time_till, contract_time, products and allow_udc will be ignored.
 
     Args:
         api_client: PowerBot ApiClient
@@ -64,11 +54,9 @@ def get_contracts(api_client: Union[ApiClient, HistoryApiClient],
         products (list): Optional list of specific products to return
         allow_udc (bool): True if user-defined contracts (block products) should also be returned on contract_time = all
         delivery_areas (list): List of EIC-codes
-        return_contract_objects (bool): If True, returns complete Contract object
 
     Returns:
-        dict{key: (list[str])}: Dictionary of Contract IDs
-        (Optional) Contract: Contract Object
+        list[ContractItem]: Dictionary of Contract Items
     """
     return get_contract_ids(api_client=api_client,
                             time_from=time_from,
@@ -89,12 +77,11 @@ def get_contract_ids(api_client: Union[ApiClient, HistoryApiClient],
                      products: list[str] = None,
                      allow_udc: bool = False,
                      delivery_areas: list[str] = None,
-                     return_contract_objects: bool = False) -> Union[dict[str, list[str]], tuple[dict, dict]]:
+                     return_contract_objects: bool = False) -> Union[dict[str, list[str]], list[ContractItem]]:
     """
-    Loads all contract IDs for specified timeframe, either by hours, single day or multiple days. Alternatively, a list
-    of contract IDs can be passed instead of a timeframe, returning a dictionary of contract IDs compatible with all other
-    functions of the Backtesting package. If return_contract_objects is True, a dictionary of contract objects will be
-    returned.
+    Loads all contract IDs for specified timeframe. Alternatively, a list of contract IDs can be passed instead of a timeframe, returning a
+    dictionary of contract IDs compatible with all other functions of the Backtesting package. If return_contract_objects is True, a list
+    of contract items will be returned.
 
     If a historic API client is used all parameters except time_from, time_till, contract_time, products and
     allow_udc will be ignored.
@@ -112,7 +99,7 @@ def get_contract_ids(api_client: Union[ApiClient, HistoryApiClient],
 
     Returns:
         dict{key: (list[str])}: Dictionary of Contract IDs
-        (Optional) Contract: Contract Object
+        OR: list[ContractItem]: Contract Object
     """
     # Check date input
     timeframes = {"all": 15, "hourly": 60, "half-hourly": 30, "quarter-hourly": 15}
@@ -183,9 +170,9 @@ def get_contract_ids(api_client: Union[ApiClient, HistoryApiClient],
     if not contracts:
         print("There was no contract data for your request!")
 
-    # Returns Contract IDs and Object As A Whole
+    # Return Contract Objects
     if return_contract_objects:
-        return contracts
+        return [i for v in contracts.values() for i in v]
 
     return contract_ids
 
@@ -202,8 +189,7 @@ def get_public_trades(api_client: Union[ApiClient, HistoryApiClient],
                       gzip_files: bool = True,
                       as_csv: bool = False) -> dict[str, pd.DataFrame]:
     """
-    Load trade data for given contract IDs. If add_vwap is True, VWAP will be calculated for each trade, incorporating
-    all previous trades.
+    Load trade data for given contract IDs. If add_vwap is True, VWAP will be calculated for each trade, incorporating all previous trades.
 
     Args:
         api_client: PowerBot ApiClient if loading from API else HistoryApiClient
@@ -232,7 +218,7 @@ def get_public_trades(api_client: Union[ApiClient, HistoryApiClient],
         # Find __cache__ directory
         cache_path = _find_cache()
 
-        for key, value in contract_ids.items():
+        for key, value in tqdm(contract_ids.items(), desc="Loading Cached Trades", unit="time periods", leave=False):
             filepath = _get_file_cachepath(api_client, key, delivery_area)
             tmp_df = None
 
@@ -458,7 +444,7 @@ def get_contract_history(api_client: Union[ApiClient, HistoryApiClient],
         contract_api = ContractApi(api_client)
         contract_ids = missing_contracts if use_cached_data else contract_ids
 
-        for key, value in contract_ids.items():
+        for key, value in tqdm(contract_ids.items(), desc="Downloading Order History", unit="time periods", leave=False):
             public_contract_history = []
             for nr, item in enumerate(value):
                 more_revisions = True
