@@ -1,18 +1,18 @@
-import asyncio
 import gzip
 import json
-import pickle
-from datetime import datetime, timedelta, timezone
+import re
+from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED
+from datetime import datetime, timezone
 from typing import Union
 
 import numpy as np
 import pandas as pd
 from powerbot_client import ApiClient, ContractApi
+from tqdm import tqdm
 
 from powerbot_backtesting.data_acquisition import get_public_trades_by_days
-from powerbot_backtesting.exceptions import NotInCacheError
 from powerbot_backtesting.models import HistoryApiClient
-from powerbot_backtesting.utils import _find_cache, _cache_data, _orderbook_data_transformation, _delta_filter
+from powerbot_backtesting.utils import _find_cache, _cache_data, _process_orderbook
 from powerbot_backtesting.utils.constants import *
 
 
@@ -31,7 +31,7 @@ def get_orders(contract_hist_data: dict[str, pd.DataFrame],
     """
     order_list = {}
 
-    for key, value in contract_hist_data.items():
+    for key, value in tqdm(contract_hist_data.items(), desc="Extracting Orders", unit="time periods", leave=False):
         value.replace(np.nan, 0, inplace=True)
         bids_asks = []
         if "orders" in value:
@@ -121,7 +121,7 @@ def get_ohlc_data(trade_data: dict[str, pd.DataFrame],
     # Check if __cache__ already exists
     cache_path = _find_cache()
 
-    for key, value in trade_data.items():
+    for key, value in tqdm(trade_data.items(), desc="Processing Trades", unit="time periods", leave=False):
         # Check If Data Already Cached
         delivery_date = datetime.strptime(key.split(" ")[0], DATE_YMD)
         year_month = delivery_date.strftime(DATE_YM)
@@ -215,10 +215,11 @@ def get_orderbooks(contract_hist_data: dict[str, pd.DataFrame],
                    api_client: Union[ApiClient, HistoryApiClient] = None,
                    use_cached_data: bool = True,
                    caching: bool = True,
-                   as_json: bool = False) -> dict[str, pd.DataFrame]:
+                   as_json: bool = False,
+                   concurrent: bool = False) -> dict[str, pd.DataFrame]:
     """
-    Converts contract history data into order books in the specified timesteps. If no API client is passed, the function
-    will automatically assume that the data is production data.
+    Converts contract history data into order books in the specified timesteps. If no API client is passed, the function will automatically assume
+    that the data is production data.
 
     Please be aware that optimally only data from one exchange at a time should be used (e.g. only EPEX).
 
@@ -236,6 +237,7 @@ def get_orderbooks(contract_hist_data: dict[str, pd.DataFrame],
         use_cached_data (bool): If True, function tries to load data from cache wherever possible
         caching (bool): True if single order books should be cached as JSON
         as_json (bool): True if complete order book should be cached as JSON
+        concurrent (bool): True if processing should be multithreaded -> possible performance gain on big datasets and good CPU
 
     Returns:
         dict{key: DataFrame}: Dictionary of DataFrames
@@ -243,6 +245,9 @@ def get_orderbooks(contract_hist_data: dict[str, pd.DataFrame],
     # Initial check
     if not contract_hist_data:
         raise ValueError("Warning: Provided order data is empty")
+    if concurrent and len(contract_hist_data) < 20:
+        print("Warning: using multithreading on a small dataset might increase processing time. Minimum recommended size: >20 contract periods\n"
+              "Defaulting to normal processing.")
 
     # Setup
     all_order_books = {}
@@ -255,11 +260,9 @@ def get_orderbooks(contract_hist_data: dict[str, pd.DataFrame],
     # Parameters
     contract_times = sorted([i for i in [*contract_hist_data]])
     first_contract = datetime.strptime(
-        contract_times[0].replace(f"{str(contract_times[0]).split(' - ')[1]}", "").replace(" - ", ":00"),
-        DATE_YMD_TIME_HMS)
+        contract_times[0].replace(f"{str(contract_times[0]).split(' - ')[1]}", "").replace(" - ", ":00"), DATE_YMD_TIME_HMS)
     last_contract = datetime.strptime(
-        contract_times[-1].replace(f"{str(contract_times[-1]).split(' ')[1]}", "").replace(" - ", "") + ":00",
-        DATE_YMD_TIME_HMS)
+        contract_times[-1].replace(f"{str(contract_times[-1]).split(' ')[1]}", "").replace(" - ", "") + ":00", DATE_YMD_TIME_HMS)
 
     year_month = first_contract.strftime(DATE_YM)
     day_month = first_contract.strftime(DATE_MD)
@@ -279,18 +282,39 @@ def get_orderbooks(contract_hist_data: dict[str, pd.DataFrame],
     cache_path = _find_cache()
     new_dir = cache_path.joinpath(f"{environment}\\{exchange}_{delivery_area}\\{year_month}\\{day_month}\\processed")
 
-    # Processing Concurrently With Asyncio
-    async def main():
-        function_list = []
-        for nr, (key, value) in enumerate(contract_hist_data.items()):
-            function_list.append(
-                __process_orderbook(key, value, str(new_dir), timesteps, time_unit,
-                                    timestamp[nr] if timestamp else None, from_timestamp,
-                                    all_order_books, use_cached_data))
+    if concurrent and len(contract_hist_data) >= 20:
+        # Define a simple function to submit to thread pool executor
+        def process_multiple(contract_hist_data):
+            for nr, (key, value) in enumerate(
+                    tqdm(contract_hist_data.items(), desc="Processing Orders (multithreading)", position=0, unit="time periods", leave=False)):
+                _process_orderbook(key, value, str(new_dir), timesteps, time_unit, timestamp[nr] if timestamp else None, from_timestamp,
+                                   all_order_books, use_cached_data)
 
-        await asyncio.gather(*function_list)
+        # Define a splitter function that splits data into appropriately sized chunks without losing data
+        def splitter(data, split):
+            counter = 0
+            out = []
+            while counter != split:
+                out.append({k: v for k, v in data.items() if k in [*data][int(len(data) / split) * counter:int(len(data) / split) * (counter + 1)]})
+                counter += 1
 
-    asyncio.run(main())
+            # Distribute rest of data
+            if len(data) % split:
+                rest = len(data) - sum([len(i) for i in out])
+                for i in range(rest):
+                    out[i] |= {[*data][-(i + 1)]: data[[*data][-(i + 1)]]}
+            return out
+
+        # Processing Concurrently with ThreadPoolExecutor & waiting for results
+        executor = ThreadPoolExecutor()
+        results = [executor.submit(process_multiple, chunk) for chunk in splitter(contract_hist_data, executor._max_workers)]
+        wait(results, return_when=ALL_COMPLETED)
+
+    else:
+        # Processing Synchronously
+        for nr, (key, value) in enumerate(tqdm(contract_hist_data.items(), desc="Processing Orders", unit="time periods", leave=False)):
+            _process_orderbook(key, value, str(new_dir), timesteps, time_unit, timestamp[nr] if timestamp else None, from_timestamp, all_order_books,
+                               use_cached_data)
 
     # Cache Data as pickle
     if caching:
@@ -310,146 +334,6 @@ def get_orderbooks(contract_hist_data: dict[str, pd.DataFrame],
     return all_order_books
 
 
-async def __process_orderbook(key: str,
-                              value: pd.DataFrame,
-                              directory: str,
-                              timesteps: int,
-                              time_unit: str,
-                              timestamp: Union[datetime, None],
-                              from_timestamp: bool,
-                              orderbook_dict: dict[str, pd.DataFrame],
-                              use_cached_data: bool):
-    """
-    Asyncio function to process single order book. Return value is appended to collection of order books.
-
-    Returns:
-        pd.Dataframe: single order book
-    """
-    # Setup Parameters
-    units = {"hours": 0, "minutes": 0, "seconds": 0, time_unit: timesteps}
-    delivery_start = datetime.strptime(key.replace(f"{str(key).split(' - ')[1]}", "").replace(" - ", ":00"),
-                                       DATE_YMD_TIME_HMS).replace(tzinfo=timezone.utc)
-    timestamp = timestamp.replace(tzinfo=timezone.utc) if timestamp else None
-    file_name = key.replace(f"{str(key).split(' ')[0]}", "").replace(":", "-")
-    directory = directory.split('\\')
-    directory = '\\'.join(directory)
-
-    try:
-        if not use_cached_data:
-            raise NotInCacheError("Not loading from cache")
-        # Check If Data Already Cached
-        order_book_clean = pickle.load(open(f"{directory}\\{file_name}_orderbook_{timesteps}{time_unit[0]}.p", "rb"))
-        orderbook_dict[key] = order_book_clean
-
-    except (NotInCacheError, FileNotFoundError):
-        # Filter out emtpy revisions
-        if "orders" in value:
-            order_filter = value.orders.map(lambda x: x if x["bid"] or x["ask"] else None)
-            value = value.loc[~value.index.isin(order_filter[order_filter.isna()].index)]
-        else:
-            value = value.loc[~(value.bids.isna()) & ~(value.asks.isna())]
-
-        # Setting Either Starting Point or Specific Timestamp
-        start_time = value.as_of.min().replace(minute=0, second=0, microsecond=0, tzinfo=timezone.utc)
-        time = timestamp if timestamp and from_timestamp \
-            else timestamp + timedelta(**{time_unit: timesteps}) if timestamp and not from_timestamp \
-            else start_time + timedelta(**{time_unit: timesteps})
-        time = start_time + timedelta(**{time_unit: timesteps}) if start_time > time else time
-
-        order_book = {}
-
-        # Transform orders
-        df_bid_asks = _orderbook_data_transformation(value)
-
-        if not df_bid_asks.empty:
-            # Create a filter to shorten dataframe
-            orders_del = set()
-
-            # Main Loop
-            while time <= delivery_start:
-                # Shorten Bids_Asks
-                if len(orders_del) != 0:
-                    df_bid_asks = df_bid_asks.loc[~df_bid_asks.order_id.isin(orders_del)]
-
-                # Create New Temporary Dataframe
-                if timestamp and not from_timestamp:
-                    df_temp = df_bid_asks.loc[df_bid_asks.as_of <= f'{timestamp}']
-
-                else:
-                    df_temp = df_bid_asks.loc[(df_bid_asks.as_of >= f'{start_time}') & (df_bid_asks.as_of <= f'{time}')]
-
-                # Delete all orders before the last delta == False
-                df_temp = _delta_filter(df_temp)
-
-                if not df_temp.empty:
-                    # Extract Order IDs for Quantity = 0 & Update Set Of Order IDs
-                    contract_ids = df_temp.contract_id.unique().tolist()
-                    orders_del.update(df_temp.loc[df_temp.quantity == 0].order_id.tolist())
-
-                    # Check For Uniformity of Contract ID
-                    if len(contract_ids) == 1:
-                        # QC For Temporary Dataframe
-                        # Add Filtered Df To Orderbook
-                        if timestamp and not from_timestamp:
-                            order_book[f"{timestamp}"] = df_temp.loc[~df_temp.order_id.isin(orders_del)]
-                        else:
-                            order_book[f"{time}"] = df_temp.loc[~df_temp.order_id.isin(orders_del)]
-
-                    else:
-                        # If There Are Multiple Contracts In The Same Orderbook -> Create 2 Separate Orderbooks
-                        dataframes = []
-                        df_check_1 = df_temp.loc[df_temp.contract_id == contract_ids[0]]
-                        dataframes.append(df_check_1)
-                        df_check_2 = df_temp.loc[df_temp.contract_id == contract_ids[1]]
-                        df_check_2 = df_check_2.loc[df_check_2.as_of > f'{time - timedelta(**units)}']
-                        dataframes.append(df_check_2)
-
-                        # Quality Control For Temporary Dataframe
-                        temp_dataframe_list = []
-
-                        for nr, val in enumerate(dataframes):
-                            # Quality Control For Temporary Dataframe
-                            df_check = val.loc[~val.order_id.isin(orders_del)]
-
-                            if not df_check.empty:
-                                temp_dataframe_list.append(df_check)  # Add Filtered Df To List
-
-                        if len(temp_dataframe_list) == 1:
-                            order_book[f"{timestamp if timestamp and not from_timestamp else time}"] = \
-                                temp_dataframe_list[0]
-                        else:
-                            for nr, val in enumerate(temp_dataframe_list):
-                                if not nr:
-                                    order_book[
-                                        f"{(timestamp if timestamp and not from_timestamp else time) - timedelta(seconds=1)}"] = val
-                                else:
-                                    order_book[f"{timestamp if timestamp and not from_timestamp else time}"] = val
-
-                # Progress In Time Or Break Loop If Timestamp Exists
-                if timestamp and not from_timestamp:
-                    break
-
-                # Adjust Start Time To New Contract if necessary
-                if time == (delivery_start - timedelta(hours=1)):
-                    start_time = time
-
-                # Time Progression
-                time += timedelta(**units)
-
-            # General Quality Control
-            # Delete All Order ID Duplicates & Empty Timesteps
-            order_book_clean = {
-                key: value.sort_values(by=['as_of'], ascending=False).drop_duplicates(subset="order_id", keep="first",
-                                                                                      inplace=False) for (key, value) in
-                order_book.items()}
-            order_book_clean = {key: value for (key, value) in order_book_clean.items() if not value.empty}
-
-        else:
-            order_book_clean = df_bid_asks
-
-    orderbook_dict[key] = order_book_clean
-
-
 def calc_trade_vwap(api_client: ApiClient,
                     contract_time: str,
                     delivery_area: str,
@@ -459,9 +343,8 @@ def calc_trade_vwap(api_client: ApiClient,
                     contract_id: str = None,
                     index: str = "ID3") -> pd.DataFrame:
     """
-    Function gets trades for a certain contract for X previous days in the same delivery period and calculates their
-    VWAP for ID3 or ID1 or all. Generates a new list of trades for these contracts.
-    Can take either a time period or a specific contract ID to load data for.
+    Function gets trades for a certain contract for X previous days in the same delivery period and calculates their VWAP for ID3 or ID1 or all.
+    Generates a new list of trades for these contracts. Can take either a time period or a specific contract ID to load data for.
 
     If previous days is 0, only the trades for the original time period/ contract will be loaded.
 
@@ -506,7 +389,7 @@ def calc_trade_vwap(api_client: ApiClient,
                                                                          contract_time=contract_time)
 
     # Processing
-    for key, value in trade_data.items():
+    for key, value in tqdm(trade_data.items(), desc="Calculating Trade VWAPs", unit="time periods", leave=False):
         # Time Difference In Minutes
         time = datetime.strptime(key.replace(f"{str(key).split(' - ')[1]}", "").replace(" - ", ":00"),
                                  DATE_YMD_TIME_HMS)
@@ -540,9 +423,9 @@ def calc_trade_vwap(api_client: ApiClient,
     return all_trade_data.reset_index(drop=True)
 
 
-def calc_single_vwap(objects: dict[str, pd.DataFrame],
-                     desired_depth: float,
-                     min_depth: float = None) -> dict[str, float]:
+def vwap_by_depth(objects: dict[str, pd.DataFrame],
+                  desired_depth: float,
+                  min_depth: float = None) -> dict[str, float]:
     """
     This method can be used to calculate the weighted average price for a dictionary of dataframes (e.g. orders, trades) at a desired depth.
     The output is a singular value for each dataframe. This function does not load any data, therefore the already existing data object has to
@@ -561,7 +444,7 @@ def calc_single_vwap(objects: dict[str, pd.DataFrame],
 
     vwaps = {k: 0 for k in [*objects]}
 
-    for key, obj in objects.items():
+    for key, obj in tqdm(objects.items(), desc="Calculating Single VWAPs", unit="time periods", leave=False):
         available_depth = 0
         total_value = 0
 
@@ -583,6 +466,56 @@ def calc_single_vwap(objects: dict[str, pd.DataFrame],
     return vwaps
 
 
+def vwap_by_timeperiod(objects: pd.DataFrame,
+                       timestamp: str,
+                       time_spec: str = "60T-60T-0T") -> float:
+    """
+    Function to calculate the value-weighted average price at the given point in time for the last X minutes.
+
+    To specify the time period precisely, the time_spec parameter should be used. The pattern is always as follows:
+
+    {60/30/15/0}T-{60/45/30/15}T-{45/30/15/0}T
+
+    Explanation:
+        {60/30/15/0}T -> Floor, will count back to the last full hour/ half-hour/ quarter-hour / last minute and act as starting point
+        {60/45/30/15}T -> Execution From, determines the minutes that should be subtracted from Floor to reach starting point for calculation
+        {45/30/15/0}T -> Execution To, determines the minutes that should be subtracted from Floor to reach end point for calculation
+
+    Examples:
+        60T-60T-0T  <--> VWAP of the previous trading hour.
+        60T-15T-0T  <--> VWAP of the last quarter-hour of the previous trading hour.
+        60T-30T-15T <--> VWAP of third quarter-hour of the previous trading hour.
+        15T-60T-0T  <--> VWAP of last hour calculated from last quarter hour.
+        0T-60T-30T  <--> VWAP of first half of the last hour calculated from current timestamp.
+
+    Args:
+        objects (pd.DataFrame): Collection of trades/ orders
+        timestamp (str): Current timestamp
+        time_spec (str): String of time specification as explained above
+
+    Returns:
+        float
+    """
+    if objects is None or isinstance(objects, pd.DataFrame) and objects.empty or not bool(
+            re.search("((60|30|15|0)T-(60|45|30|15)T-(45|30|15|0)T)", time_spec)):
+        return 0
+
+    # Parse time specification
+    timestamp = pd.Timestamp(timestamp, tz="utc")
+    time_periods = [int(t) for t in time_spec.replace("T", "").split("-")]
+    recalculation_point = timestamp.floor(freq=f"{time_periods[0] or ''}T")
+    execution_from = recalculation_point - pd.Timedelta(minutes=time_periods[1])
+    execution_to = recalculation_point - pd.Timedelta(minutes=time_periods[2])
+
+    # Filter and check dataframe
+    field = "exec_time" if "exec_time" in objects.columns else "as_of"
+
+    if (filtered := objects.loc[(objects[field] >= execution_from) & (objects[field] < execution_to)]).empty:
+        return 0
+
+    return round(sum(values["price"] * values["quantity"] for row, values in filtered.iterrows()) / sum(filtered.quantity), 2)
+
+
 def calc_rolling_vwap(trades: dict[str, pd.DataFrame],
                       rolling_interval: int = 1,
                       rolling_time_unit: str = "hour") -> dict[str, pd.DataFrame]:
@@ -601,7 +534,7 @@ def calc_rolling_vwap(trades: dict[str, pd.DataFrame],
     """
     t_map = {"hour": "h", "minute": "min", "second": "s"}
 
-    for k, v in trades.items():
+    for k, v in tqdm(trades.items(), desc="Calculating Rolling VWAPs", unit="time periods", leave=False):
         t = f"{rolling_interval}{t_map[rolling_time_unit]}"
 
         # Set index on temporary dataframe
@@ -611,10 +544,8 @@ def calc_rolling_vwap(trades: dict[str, pd.DataFrame],
         ptq = v_temp.price * v_temp.quantity
         quant_sum = v_temp.rolling(window=t)["quantity"].sum()
         ptq_sum = ptq.rolling(window=t).sum()
-        temp = ptq_sum / quant_sum
-        temp = [round(i, 2) for i in temp.to_list()]
 
         # Add to original df
-        v[f"vwap_{t}"] = temp
+        v[f"vwap_{t}"] = round(ptq_sum / quant_sum, 2).to_list()
 
     return trades
