@@ -257,10 +257,10 @@ def get_historic_contract_ids(client: HistoryApiClient,
                               contract_time: str,
                               products: list[str],
                               allow_udc: bool,
-                              return_contract_objects: bool = False) -> Union[dict[str, list[str], list[dict]]]:
+                              return_contract_objects: bool = False) -> Union[dict[str, list[str]], list[dict]]:
     """
-    Function mimics get_contract_ids but loads contract IDs from a cached index file instead of the API. This is
-    necessary when trying to load data that is older than the data retention policy for production instances.
+    Function mimics get_contract_ids but loads contract IDs from a cached index file instead of the API. This is necessary when trying to load data
+    that is older than the data retention policy for production instances.
 
     Args:
         client: HistoryApiClient
@@ -273,6 +273,7 @@ def get_historic_contract_ids(client: HistoryApiClient,
 
     Returns:
         dict{key: (list[str])}: Dictionary of Contract IDs or List of Dictionaries
+        OR: list[ContractItem]: Contract Object
     """
     # Setup
     timeframes = {"all": 15, "hourly": 60, "half-hourly": 30, "quarter-hourly": 15}
