@@ -6,6 +6,7 @@ from pathlib import Path
 from typing import Union
 
 import pandas as pd
+from tqdm import tqdm
 
 from powerbot_backtesting.utils import _cache_data, _find_cache, _historic_data_transformation, _historic_contract_transformation
 from powerbot_backtesting.utils.constants import *
@@ -71,7 +72,7 @@ def process_historic_data(extracted_files: Union[list[str], dict[str, list[str]]
                     v.contract_id]
 
         # For each timestep
-        for time, ids in contract_times.items():
+        for time, ids in tqdm(contract_times.items(), desc="Historic File Transformation", unit="files", leave=False):
             files = {"trades": [i for i in unzipped_files if any(str(x) in i for x in ids) and "Trades" in i],
                      "orders": [i for i in unzipped_files if any(str(x) in i for x in ids) and "Orders" in i]}
 
