@@ -10,15 +10,18 @@ from urllib.parse import quote
 
 import pandas as pd
 import plotly.graph_objects as go
+from dateutil.tz import tzutc
 from powerbot_client import Signal, Trade, InternalTrade, OwnOrder
 from pydantic import BaseModel, Field, validate_arguments, root_validator, validator
 from sqlalchemy import create_engine
 from sqlalchemy.exc import NoSuchModuleError, InterfaceError, ProgrammingError, OperationalError, TimeoutError
 from sqlalchemy.orm import Session
+from tqdm import tqdm
 
 import powerbot_backtesting as pb
 from powerbot_backtesting.exceptions import SQLExporterError
 from powerbot_backtesting.utils.constants import *
+from powerbot_backtesting.utils.helper_functions import _cache_data, _find_cache, _get_file_cachepath
 
 # Implement custom type for validation
 pandas_DataFrame = TypeVar('pandas.core.frame.DataFrame')
@@ -96,7 +99,8 @@ class BaseExporter(BaseModel, ABC):
                        from_timestamp: bool = False,
                        use_cached_data: bool = True,
                        caching: bool = True,
-                       as_json: bool = False) -> dict[str, pd.DataFrame]:
+                       as_json: bool = False,
+                       concurrent: bool = False) -> dict[str, pd.DataFrame]:
 
         """
         Converts contract history data into order books in the specified timesteps. If no API client is passed, the function
@@ -117,6 +121,7 @@ class BaseExporter(BaseModel, ABC):
             use_cached_data (bool): If True, function tries to load data from cache wherever possible
             caching (bool): True if single order books should be cached as JSON
             as_json (bool): True if complete order book should be cached as JSON
+            concurrent (bool): True if processing should be multithreaded -> possible performance gain on big datasets and good CPU
 
         Returns:
             dict{key: DataFrame}: Dictionary of DataFrames
@@ -133,7 +138,8 @@ class BaseExporter(BaseModel, ABC):
                                  api_client=self.client if hasattr(self, 'client') else None,
                                  use_cached_data=use_cached_data,
                                  caching=caching,
-                                 as_json=as_json)
+                                 as_json=as_json,
+                                 concurrent=concurrent)
 
     @staticmethod
     @validate_arguments
@@ -155,9 +161,9 @@ class BaseExporter(BaseModel, ABC):
 
     @staticmethod
     @validate_arguments
-    def calc_single_vwap(objects: dict[str, pandas_DataFrame],
-                         desired_depth: float,
-                         min_depth: float = None) -> dict[str, float]:
+    def vwap_by_depth(objects: dict[str, pandas_DataFrame],
+                      desired_depth: float,
+                      min_depth: float = None) -> dict[str, float]:
         """
         This method can be used to calculate the weighted average price for a dictionary of dataframes (e.g. orders, trades) at a desired depth.
         The output is a singular value for each dataframe. This function does not load any data, therefore the already existing data object has to
@@ -171,9 +177,45 @@ class BaseExporter(BaseModel, ABC):
         Returns:
             dict[str, float]: The weighted average price for the desired depth for each key in the dictionary.
         """
-        return pb.calc_single_vwap(objects=objects,
-                                   desired_depth=desired_depth,
-                                   min_depth=min_depth)
+        return pb.vwap_by_depth(objects=objects,
+                                desired_depth=desired_depth,
+                                min_depth=min_depth)
+
+    @staticmethod
+    @validate_arguments
+    def vwap_by_timeperiod(objects: pandas_DataFrame,
+                           timestamp: str,
+                           time_spec: str = "60T-60T-0T") -> float:
+        """
+        Function to calculate the value-weighted average price at the given point in time for the last X minutes.
+
+        To specify the time period precisely, the time_spec parameter should be used. The pattern is always as follows:
+
+        {60/30/15/0}T-{60/45/30/15}T-{45/30/15/0}T
+
+        Explanation:
+            {60/30/15/0}T -> Floor, will count back to the last full hour/ half-hour/ quarter-hour / last minute and act as starting point
+            {60/45/30/15}T -> Execution From, determines the minutes that should be subtracted from Floor to reach starting point for calculation
+            {45/30/15/0}T -> Execution To, determines the minutes that should be subtracted from Floor to reach end point for calculation
+
+        Examples:
+            60T-60T-0T  <--> VWAP of the previous trading hour.
+            60T-15T-0T  <--> VWAP of the last quarter-hour of the previous trading hour.
+            60T-30T-15T <--> VWAP of third quarter-hour of the previous trading hour.
+            15T-60T-0T  <--> VWAP of last hour calculated from last quarter hour.
+            0T-60T-30T  <--> VWAP of first half of the last hour calculated from current timestamp.
+
+        Args:
+            objects (pd.DataFrame): Collection of trades/ orders
+            timestamp (str): Current timestamp
+            time_spec (str): String of time specification as explained above
+
+        Returns:
+            float
+        """
+        return pb.vwap_by_timeperiod(objects=objects,
+                                     timestamp=timestamp,
+                                     time_spec=time_spec)
 
     @staticmethod
     @validate_arguments
@@ -321,7 +363,22 @@ class ApiExporter(BaseExporter):
                       contract_time: str = "all",
                       products: list[str] = None,
                       allow_udc: bool = False,
-                      delivery_areas: list[str] = None) -> list:
+                      delivery_areas: list[str] = None) -> list[dict]:
+        """
+        Loads all contracts for specified timeframe. Alternatively, a list of contract IDs can be passed instead of a timeframe.
+
+        Args:
+            time_from (datetime): yyyy-mm-dd hh:mm:ss
+            time_till (datetime): yyyy-mm-dd hh:mm:ss
+            contract_ids (list[str]): Optionally, a list of specific contract IDs can be passed to return contract objects
+            contract_time (str): all, hourly, half-hourly or quarter-hourly (all includes UDC)
+            products (list): Optional list of specific products to return
+            allow_udc (bool): True if user-defined contracts (block products) should also be returned on contract_time = all
+            delivery_areas (list): List of EIC-codes
+
+        Returns:
+            dict{key: (list[str])}: Dictionary of Lists containing Contracts
+        """
         return pb.get_contracts(api_client=self.client,
                                 time_from=time_from,
                                 time_till=time_till,
@@ -339,18 +396,10 @@ class ApiExporter(BaseExporter):
                          contract_time: str = "all",
                          products: list[str] = None,
                          allow_udc: bool = False,
-                         delivery_areas: list[str] = None) -> Union[dict[str, list[str]], tuple[dict, dict]]:
+                         delivery_areas: list[str] = None) -> dict[str, list[str]]:
         """
-        Loads all contract IDs for specified timeframe, either by hours, single day or multiple days. Alternatively, a list
-        of contract IDs can be passed instead of a timeframe, returning a dictionary of contract IDs compatible with all other
-        functions of the Backtesting package. If return_contract_objects is True, a dictionary of contract objects will be
-        returned as well.
-
-        If time_from includes hh:mm:ss
-            loads data between both times in timesteps according to contract_time
-
-        If time_from doesn't include hh:mm:ss
-            loads data between both dates in timesteps according to contract_time
+        Loads all contract IDs for specified timeframe. Alternatively, a list of contract IDs can be passed instead of a timeframe, returning a
+        dictionary of contract IDs compatible with all other functions of the Backtesting package.
 
         Args:
             time_from (datetime): yyyy-mm-dd hh:mm:ss
@@ -360,11 +409,9 @@ class ApiExporter(BaseExporter):
             products (list): Optional list of specific products to return
             allow_udc (bool): True if user-defined contracts (block products) should also be returned on contract_time = all
             delivery_areas (list): List of EIC-codes
-            return_contract_objects (bool): If True, returns complete Contract object
 
         Returns:
             dict{key: (list[str])}: Dictionary of Contract IDs
-            (Optional) Contract: Contract Object
         """
         return pb.get_contract_ids(api_client=self.client,
                                    time_from=time_from,
@@ -698,17 +745,10 @@ class HistoryExporter(BaseExporter):
                       time_till: datetime = Field(description="Datetime in format yyyy-mm-dd hh:mm:ss"),
                       contract_time: Optional[str] = "all",
                       products: Optional[list[str]] = None,
-                      allow_udc: Optional[bool] = False) -> dict[str, list[str]]:
+                      allow_udc: Optional[bool] = False) -> list[dict]:
         """
-        Loads all contract IDs for specified timeframe, either by hours, single day or multiple days from the local cache.
-        The cached data has to have been loaded via get_historic_data, since this function utilizes the contract file
-        as an index.
-
-        If time_from includes hh:mm:ss
-            loads data between both times in timesteps according to contract_time
-
-        If time_from doesn't include hh:mm:ss
-            loads data between both dates in timesteps according to contract_time
+        Loads all contract IDs for specified timeframe from the local cache. The cached data has to have been loaded via get_historic_data, as
+        this function utilizes the contract file as an index.
 
         Args:
             time_from (datetime): yyyy-mm-dd hh:mm:ss
@@ -718,7 +758,7 @@ class HistoryExporter(BaseExporter):
             allow_udc (bool): True if user-defined contracts (block products) should also be returned on contract_time = all
 
         Returns:
-            dict{key: (list[str])}: Dictionary of Contract IDs
+            list[dict]: List of Contract Dictionaries
         """
         return pb.get_historic_contract_ids(client=self.client,
                                             time_from=time_from,
@@ -736,15 +776,8 @@ class HistoryExporter(BaseExporter):
                          products: Optional[list[str]] = None,
                          allow_udc: Optional[bool] = False) -> dict[str, list[str]]:
         """
-        Loads all contract IDs for specified timeframe, either by hours, single day or multiple days from the local cache.
-        The cached data has to have been loaded via get_historic_data, since this function utilizes the contract file
-        as an index.
-
-        If time_from includes hh:mm:ss
-            loads data between both times in timesteps according to contract_time
-
-        If time_from doesn't include hh:mm:ss
-            loads data between both dates in timesteps according to contract_time
+        Loads all contract IDs for specified timeframe from the local cache. The cached data has to have been loaded via get_historic_data, as this
+        function utilizes the contract file as an index.
 
         Args:
             time_from (datetime): yyyy-mm-dd hh:mm:ss
@@ -1059,6 +1092,13 @@ class SQLExporter(BaseExporter):
         Returns:
             dict{key: (list[str])}: Dictionary of Contract IDs
         """
+        del_units = {"hourly": 1, "half-hourly": 0.5, "quarter-hourly": 0.25}
+
+        if (del_units := del_units.get(contract_time)):
+            sql_addition = f" AND delivery_units = {del_units}"
+        else:
+            sql_addition = ""
+
         if not isinstance(time_from, datetime) or not isinstance(time_till, datetime):
             raise SQLExporterError("Please use datetime format for time_from & time_till!")
 
@@ -1068,13 +1108,13 @@ class SQLExporter(BaseExporter):
                                      f"AND delivery_end <= '{time_till}' "
                                      f"AND delivery_areas LIKE '%{delivery_area}%' "
                                      f"AND product IN {PRODUCTS[contract_time]} "
-                                     f"AND exchange = '{exchange}'").fetchall()
+                                     f"AND exchange = '{exchange}'" + sql_addition).fetchall()
             if not result:
                 result = session.execute(f"SELECT delivery_start, delivery_end, contract_id FROM contracts "
                                          f"WHERE delivery_start >= '{time_from}' "
                                          f"AND delivery_end <= '{time_till}' "
                                          f"AND product IN {PRODUCTS[contract_time]} "
-                                         f"AND exchange = '{exchange}'").fetchall()
+                                         f"AND exchange = '{exchange}'" + sql_addition).fetchall()
 
         if not as_list:
             contract_ids = {f"{i[0].strftime(DATE_YMD_TIME_HM)} - {i[1].strftime(TIME_HM)}": [] for i in result}
@@ -1094,6 +1134,11 @@ class SQLExporter(BaseExporter):
     def get_public_trades(self,
                           as_dataframe: bool = True,
                           delivery_area: list[str] = None,
+                          exchange: str = "epex",
+                          use_cached_data: bool = True,
+                          caching: bool = False,
+                          gzip_files: bool = True,
+                          as_csv: bool = False,
                           **kwargs) -> Union[pd.DataFrame, dict[str, pd.DataFrame]]:
         """
         Exports trades from SQL database. To use different mathematical/SQL operators, pass keyworded arguments
@@ -1107,40 +1152,86 @@ class SQLExporter(BaseExporter):
 
         Args:
             as_dataframe (bool): If False -> returns list
-            delivery_area (tuple[str]): Multiple delivery areas inside a tuple. Single del. area can be passed as a string
+            delivery_area (list): List of EIC Area Codes for Delivery Areas
+            exchange (str): Exchange of requested data
+            use_cached_data (bool): If True, function tries to load data from cache wherever possible
+            caching (bool): True if data should be cached
+            gzip_files (bool): True if cached files should be gzipped
+            as_csv (bool): if True, will save files as CSV, additionally to JSON
             **kwargs: any additional fields of SQL table
 
         Returns:
             list/ DataFrame: SQL query
         """
-        allowed_kwargs = ["price", "quantity", "prc_x_qty", "exchange", "contract_id", "trade_id", "exec_time",
-                          "api_timestamp", "self_trade"]
+        if as_dataframe and len(delivery_area) > 1:
+            raise ValueError("Please only give one delivery area when loading data as dataframes")
 
-        sql_params = self.__handle_sql_args(kwargs, allowed_kwargs)
-        if delivery_area:
-            for i in delivery_area:
-                sql_params += f" AND (buy_delivery_area = '{i}' OR sell_delivery_area = '{i}')"
+        if use_cached_data and as_dataframe and (contract_ids := kwargs.get("contract_id")):
+            trades = {}
+            missing_contracts = {}
 
-        with self.__get_session() as session:
-            result = session.execute(f"SELECT * FROM public_trades{sql_params}").fetchall()
+            # Find __cache__ directory
+            cache_path = _find_cache()
+
+            for key, value in tqdm(contract_ids.items(), desc="Loading Cached Trades", unit="time periods", leave=False):
+                filepath = _get_file_cachepath(None, key, delivery_area[0], exchange)
+                tmp_df = None
+
+                for i in [".json.gz", ".json"]:
+                    if cache_path.joinpath(f"{filepath}_trades{i}").exists():
+                        tmp_df = pd.read_json(cache_path.joinpath(f"{filepath}_trades{i}"), dtype=False)
+
+                if isinstance(tmp_df, pd.DataFrame):
+                    tmp_df['api_timestamp'] = pd.to_datetime(tmp_df['api_timestamp'])
+                    tmp_df['exec_time'] = pd.to_datetime(tmp_df['exec_time'])
+                    tmp_df = tmp_df.astype({"price": "float64", "trade_id": "str", "contract_id": "str"})
+                    for i in ["price", "quantity"]:
+                        tmp_df[i] = round(tmp_df[i], 2)
+
+                    # Filter out any contract IDs that are not in value
+                    trades[key] = tmp_df.loc[tmp_df.contract_id.isin(value)]
+
+                else:
+                    # Save Missing Contract IDs
+                    missing_contracts[key] = value
+
+            kwargs["contract_id"] = missing_contracts if missing_contracts else contract_ids
+
+        if (contract_ids := kwargs.get("contract_id")) is None or isinstance(contract_ids, dict) and len(contract_ids) > 0:
+            allowed_kwargs = ["price", "quantity", "prc_x_qty", "exchange", "contract_id", "trade_id", "exec_time",
+                              "api_timestamp", "self_trade"]
+
+            sql_params = self.__handle_sql_args(kwargs, allowed_kwargs)
+
+            if delivery_area:
+                for i in delivery_area:
+                    sql_params += f" AND (buy_delivery_area = '{i}' OR sell_delivery_area = '{i}')"
+
+            with self.__get_session() as session:
+                if not (result := session.execute(f"SELECT * FROM public_trades{sql_params}").fetchall()):
+                    raise SQLExporterError("There is no trade data for the specified timeframe!")
+        else:
+            result = None
 
         if as_dataframe:
-            output = pd.DataFrame(result)
-            if not output.empty:
-                output = output.rename(columns={0: 'exchange', 1: 'contract_id', 2: 'trade_id', 3: 'api_timestamp',
-                                                4: 'exec_time', 5: 'buy_delivery_area', 6: 'sell_delivery_area',
-                                                7: 'price', 8: 'quantity', 9: 'prc_x_qty', 10: "currency",
-                                                11: 'self_trade'})
-                if not delivery_area:
-                    raise SQLExporterError("Please specify at least one delivery area!")
+            if result:
+                output = pd.DataFrame(result)
+                if not output.empty:
+                    output = output.rename(columns={0: 'exchange', 1: 'contract_id', 2: 'trade_id', 3: 'api_timestamp',
+                                                    4: 'exec_time', 5: 'buy_delivery_area', 6: 'sell_delivery_area',
+                                                    7: 'price', 8: 'quantity', 9: 'prc_x_qty', 10: "currency",
+                                                    11: 'self_trade'})
 
-                output['api_timestamp'] = pd.to_datetime(output['api_timestamp'], utc=True)
-                output['exec_time'] = pd.to_datetime(output['exec_time'], utc=True)
+                    output['api_timestamp'] = pd.to_datetime(output['api_timestamp'], utc=True)
+                    output['exec_time'] = pd.to_datetime(output['exec_time'], utc=True)
+                    trades |= self.__convert_dataframe("trades", output)
 
-                self.logger.info("Successfully exported trades")
-                return self.__convert_dataframe("trades", output)
-            else:
-                raise SQLExporterError("There is no trade data for the specified timeframe!")
+            # Caching
+            if caching:
+                _cache_data("trades", trades, delivery_area[0], api_client=None, gzip_files=gzip_files, as_csv=as_csv, exchange=exchange)
+
+            self.logger.info("Successfully exported trades")
+            return trades
 
         self.logger.info("Successfully exported trades")
 
@@ -1149,6 +1240,12 @@ class SQLExporter(BaseExporter):
     @validate_arguments
     def get_contract_history(self,
                              as_dataframe: bool = True,
+                             delivery_area: list[str] = None,
+                             exchange: str = "epex",
+                             use_cached_data: bool = True,
+                             caching: bool = False,
+                             gzip_files: bool = True,
+                             as_csv: bool = False,
                              **kwargs) -> Union[pd.DataFrame, dict[str, pd.DataFrame]]:
         """
         Exports contract revisions from SQL database. To use different mathematical/SQL operators, pass keyworded arguments
@@ -1163,34 +1260,109 @@ class SQLExporter(BaseExporter):
 
         Args:
             as_dataframe (bool): If False -> returns list
+            delivery_area (list): List of EIC Area Codes for Delivery Areas
+            exchange (str): Exchange of requested data
+            use_cached_data (bool): If True, function tries to load data from cache wherever possible
+            caching (bool): True if data should be cached
+            gzip_files (bool): True if cached files should be gzipped
+            as_csv (bool): if True, will save files as CSV, additionally to JSON
             **kwargs: any additional fields of SQL table
 
         Returns:
             list/ DataFrame: SQL query
         """
-        allowed_kwargs = ["exchange", "contract_id", "delivery_area", "revision_no", "as_of", "best_bid",
-                          "best_bid_qty", "best_ask", "best_ask_qty", "vwap", "high", "low", "last_price",
-                          "last_qty", "last_trade_time", "volume", "delta"]
+        orders = {}
+        missing_contracts = {}
 
-        sql_params = self.__handle_sql_args(kwargs, allowed_kwargs)
+        if as_dataframe and len(delivery_area) > 1:
+            raise ValueError("Please only give one delivery area when loading data as dataframes")
 
-        with self.__get_session() as session:
-            result = session.execute(f"SELECT * FROM contract_revisions{sql_params}").fetchall()
+        if use_cached_data and as_dataframe and (contract_ids := kwargs.get("contract_id")):
+            # Find __cache__ directory
+            cache_path = _find_cache()
+            for key, value in contract_ids.items():
+                filepath = _get_file_cachepath(None, key, delivery_area[0], exchange)
+                tmp_df = None
+
+                for i in [".json.gz", ".json"]:
+                    if cache_path.joinpath(f"{filepath}_ordhist{i}").exists():
+                        tmp_df = pd.read_json(cache_path.joinpath(f"{filepath}_ordhist{i}"), dtype=False,
+                                              convert_dates=False)
+
+                if isinstance(tmp_df, pd.DataFrame):
+                    tmp_df['as_of'] = pd.to_datetime(tmp_df['as_of'])
+
+                    cols = {"internal_trades": "object", "contract_id": "str", "auction_price": "float64"}
+                    cols = {k: v for k, v in cols.items() if k in tmp_df.columns}
+                    tmp_df = tmp_df.astype(cols, errors='ignore')
+                    for i in ["best_bid_price", "best_bid_quantity", "best_ask_price", "best_ask_quantity", "last_price",
+                              "last_quantity", "total_quantity", "high", "low", "vwap"]:
+                        try:
+                            tmp_df[i] = round(tmp_df[i], 2)
+                        except (TypeError, KeyError):
+                            pass
+
+                    if "orders" in tmp_df.columns:
+                        order_list = tmp_df.orders.tolist()
+                        for order_type in ["bid", "ask"]:
+                            for i in order_list:
+                                if order_type in i and i[order_type]:
+                                    for x in i[order_type]:
+                                        for param in ["quantity", "price"]:
+                                            x[param] = round(x[param], 2)
+                                        try:
+                                            x["order_entry_time"] = datetime.strptime(x["order_entry_time"],
+                                                                                      "%Y-%m-%dT%H:%M:%S.%fZ").replace(
+                                                tzinfo=tzutc())
+                                        except ValueError:
+                                            x["order_entry_time"] = datetime.strptime(x["order_entry_time"],
+                                                                                      "%Y-%m-%dT%H:%M:%SZ").replace(
+                                                microsecond=0, tzinfo=tzutc())
+                        tmp_df["orders"] = order_list
+
+                    # Filter out any contract IDs that are not in value
+                    orders[key] = tmp_df.loc[tmp_df.contract_id.isin(value)]
+                else:
+                    # Save Missing Contract IDs
+                    missing_contracts[key] = value
+
+            kwargs["contract_id"] = missing_contracts
+
+        if (contract_ids := kwargs.get("contract_id")) is None or isinstance(contract_ids, dict) and len(contract_ids) > 0:
+            allowed_kwargs = ["exchange", "contract_id", "delivery_area", "revision_no", "as_of", "best_bid",
+                              "best_bid_qty", "best_ask", "best_ask_qty", "vwap", "high", "low", "last_price",
+                              "last_qty", "last_trade_time", "volume", "delta"]
+
+            sql_params = self.__handle_sql_args(kwargs, allowed_kwargs)
+
+            if delivery_area and not as_dataframe and len(delivery_area) > 1:
+                sql_params += f" AND delivery_area IN {tuple(delivery_area)}"
+            else:
+                sql_params += f" AND delivery_area = '{delivery_area[0]}'"
+
+            with self.__get_session() as session:
+                if not (result := session.execute(f"SELECT * FROM contract_revisions{sql_params}").fetchall()):
+                    raise SQLExporterError("There is no order data for the specified timeframe!")
+        else:
+            result = None
 
         if as_dataframe:
-            output = pd.DataFrame(result)
-            if not output.empty:
-                output = output.rename(columns={0: 'exchange', 1: 'contract_id', 2: 'delivery_area', 3: 'revision_no',
-                                                4: 'as_of', 5: 'best_bid', 6: 'best_bid_qty', 7: 'best_ask',
-                                                8: 'best_ask_qty', 9: 'vwap', 10: 'high', 11: 'low', 12: 'last_price',
-                                                13: 'last_qty', 14: "last_trade_time", 15: 'volume', 16: 'delta',
-                                                17: 'bids', 18: 'asks'})
-                if "delivery_area" not in kwargs:
-                    raise SQLExporterError("Please specify one specific delivery area!")
-                self.logger.info("Successfully exported contract history")
-                return self.__convert_dataframe("orders", output)
-            else:
-                raise SQLExporterError("There is no order data for the specified timeframe!")
+            if result:
+                output = pd.DataFrame(result)
+                if not output.empty:
+                    output = output.rename(columns={0: 'exchange', 1: 'contract_id', 2: 'delivery_area', 3: 'revision_no',
+                                                    4: 'as_of', 5: 'best_bid', 6: 'best_bid_qty', 7: 'best_ask',
+                                                    8: 'best_ask_qty', 9: 'vwap', 10: 'high', 11: 'low', 12: 'last_price',
+                                                    13: 'last_qty', 14: "last_trade_time", 15: 'volume', 16: 'delta',
+                                                    17: 'bids', 18: 'asks'})
+                orders |= self.__convert_dataframe("orders", output)
+
+            # Caching
+            if caching:
+                _cache_data("ordhist", orders, delivery_area[0], api_client=None, gzip_files=gzip_files, as_csv=as_csv, exchange=exchange)
+
+            self.logger.info("Successfully exported contract history")
+            return orders
 
         self.logger.info("Successfully exported contract history")
 
@@ -1508,15 +1680,17 @@ class SQLExporter(BaseExporter):
                         else:
                             sql_params += f" OR {keyword} LIKE '%{element}%'"
                     continue
-
                 elif isinstance(argument, dict):
                     op = "IN"
-                    temp_list = []
-                    for value in argument.values():
-                        for item in value:
-                            temp_list.append(item)
+                    temp_list = [i for x in argument.values() for i in x]
                     argument = tuple(temp_list)
+                    if len(argument) == 1:
+                        argument = str(argument[0])
+                        op = "="
                 try:
+                    if keyword == "contract_id" and isinstance(argument, str):
+                        raise Exception
+
                     if not isinstance(argument, tuple) and keyword != "contract_id":
                         argument = float(argument)
                     sql_params += f" {sql_statement} {keyword} {op} {argument}"
@@ -1550,12 +1724,6 @@ class SQLExporter(BaseExporter):
             dataframe["bids"] = [json.loads(i) if i else None for i in dataframe.bids.tolist()]
             dataframe["asks"] = [json.loads(i) if i else None for i in dataframe.asks.tolist()]
 
-        # 	dataframe.drop(columns=["bids", "asks"])
-        # 	orders = []
-        # 	for nr, val in enumerate(bids):
-        # 		orders.append({"bid": val, "ask": asks[nr]})
-        # 	dataframe["orders"] = orders
-
         for row_nr, row_id in enumerate(contracts.contract_id):
             key = f"{contracts.iloc[row_nr].delivery_start.strftime(DATE_YMD_TIME_HM)} - " \
                   f"{contracts.iloc[row_nr].delivery_end.strftime(TIME_HM)}"
