@@ -1,6 +1,6 @@
 import os
 import pickle
-from datetime import datetime, timezone
+from datetime import datetime, timezone, timedelta
 from decimal import Decimal, getcontext
 from pathlib import Path
 from time import sleep
@@ -12,7 +12,7 @@ from powerbot_client import ApiClient, TradesApi, SignalsApi, Signal, Trade, Int
 
 from powerbot_backtesting.utils.constants import *
 from powerbot_backtesting.models import HistoryApiClient
-
+from powerbot_backtesting.exceptions import NotInCacheError
 
 def _get_private_data(api_client: ApiClient,
 					  data_type: str,
@@ -186,7 +186,7 @@ def _find_cache() -> Path:
 	return cache_path
 
 
-def _get_file_cachepath(api_client: Union[ApiClient, HistoryApiClient], contract_key: str, delivery_area: str) -> str:
+def _get_file_cachepath(api_client: Union[ApiClient, HistoryApiClient], contract_key: str, delivery_area: str, exchange: str = None) -> str:
 	"""
 	Helper function that constructs most of the path of a cached file.
 
@@ -194,13 +194,14 @@ def _get_file_cachepath(api_client: Union[ApiClient, HistoryApiClient], contract
 		api_client: PowerBot ApiClient if loading from API else HistoryApiClient
 		contract_key (str): Key of dictionary
 		delivery_area (str): EIC Area Code for Delivery Area
+		exchange (str): exchange of contracts -> needed when loading with SQLExporter
 
 	Returns:
 		filepath: str
 	"""
 	environment = api_client.configuration.host.split("/")[2].split(".")[0] if isinstance(api_client, ApiClient) else None
 	environment = "staging" if environment == "staging" else "prod"
-	market = api_client.configuration.host.split("/")[4] if isinstance(api_client, ApiClient) else api_client.exchange
+	market = api_client.configuration.host.split("/")[4] if isinstance(api_client, ApiClient) else api_client.exchange if api_client else exchange
 	delivery_date = datetime.strptime(contract_key.split(" ")[0], DATE_YMD)
 	year_month = delivery_date.strftime(DATE_YM)
 	day_month = delivery_date.strftime(DATE_MD)
@@ -215,8 +216,8 @@ def _check_contracts(contract, delivery_areas: list[str], products: list[str], a
 
 	Args:
 		contract: Contract Object
-		delivery_areas: List of EIC-codes
-		products: List of products
+		delivery_areas (list): List of EIC-codes
+		products (list): List of products
 
 	Returns:
 		bool
@@ -243,6 +244,146 @@ def _check_contracts(contract, delivery_areas: list[str], products: list[str], a
 	return True
 
 
+def _process_orderbook(key: str,
+                       value: pd.DataFrame,
+                       directory: str,
+                       timesteps: int,
+                       time_unit: str,
+                       timestamp: Union[datetime, None],
+                       from_timestamp: bool,
+                       orderbook_dict: dict[str, pd.DataFrame],
+                       use_cached_data: bool):
+    """
+    Function to process single order book. Return value is appended to collection of order books.
+
+    Returns:
+        pd.Dataframe: single order book
+    """
+    # Setup Parameters
+    units = {"hours": 0, "minutes": 0, "seconds": 0, time_unit: timesteps}
+    delivery_start = datetime.strptime(key.replace(f"{str(key).split(' - ')[1]}", "").replace(" - ", ":00"),
+                                       DATE_YMD_TIME_HMS).replace(tzinfo=timezone.utc)
+    timestamp = timestamp.replace(tzinfo=timezone.utc) if timestamp else None
+    file_name = key.replace(f"{str(key).split(' ')[0]}", "").replace(":", "-")
+    directory = directory.split('\\')
+    directory = '\\'.join(directory)
+
+    try:
+        if not use_cached_data:
+            raise NotInCacheError("Not loading from cache")
+        # Check If Data Already Cached
+        order_book_clean = pickle.load(open(f"{directory}\\{file_name}_orderbook_{timesteps}{time_unit[0]}.p", "rb"))
+        orderbook_dict[key] = order_book_clean
+
+    except (NotInCacheError, FileNotFoundError):
+        # Filter out emtpy revisions
+        if "orders" in value:
+            order_filter = value.orders.map(lambda x: x if x["bid"] or x["ask"] else None)
+            value = value.loc[~value.index.isin(order_filter[order_filter.isna()].index)]
+        else:
+            value = value.loc[~(value.bids.isna()) & ~(value.asks.isna())]
+
+        # Setting Either Starting Point or Specific Timestamp
+        start_time = value.as_of.min().replace(minute=0, second=0, microsecond=0, tzinfo=timezone.utc)
+        time = timestamp if timestamp and from_timestamp \
+            else timestamp + timedelta(**{time_unit: timesteps}) if timestamp and not from_timestamp \
+            else start_time + timedelta(**{time_unit: timesteps})
+        time = start_time + timedelta(**{time_unit: timesteps}) if start_time > time else time
+
+        order_book = {}
+
+        # Transform orders
+        df_bid_asks = _orderbook_data_transformation(value)
+
+        if not df_bid_asks.empty:
+            # Create a filter to shorten dataframe
+            orders_del = set()
+
+            # Main Loop
+            while time <= delivery_start:
+                # Shorten Bids_Asks
+                if len(orders_del) > 0:
+                    df_bid_asks = df_bid_asks.loc[~df_bid_asks.order_id.isin(orders_del)]
+
+                # Create New Temporary Dataframe
+                if timestamp and not from_timestamp:
+                    df_temp = df_bid_asks.loc[df_bid_asks.as_of <= f'{timestamp}']
+
+                else:
+                    df_temp = df_bid_asks.loc[(df_bid_asks.as_of >= f'{start_time}') & (df_bid_asks.as_of <= f'{time}')]
+
+                # Delete all orders before the last delta == False
+                df_temp = _delta_filter(df_temp)
+
+                if not df_temp.empty:
+                    # Extract Order IDs for Quantity = 0 & Update Set Of Order IDs
+                    contract_ids = df_temp.contract_id.unique().tolist()
+                    orders_del.update(df_temp.loc[df_temp.quantity == 0].order_id.tolist())
+
+                    # Check For Uniformity of Contract ID
+                    if len(contract_ids) == 1:
+                        # QC For Temporary Dataframe
+                        # Add Filtered Df To Orderbook
+                        if timestamp and not from_timestamp:
+                            order_book[f"{timestamp}"] = df_temp.loc[~df_temp.order_id.isin(orders_del)]
+                        else:
+                            order_book[f"{time}"] = df_temp.loc[~df_temp.order_id.isin(orders_del)]
+
+                    else:
+                        # If There Are Multiple Contracts In The Same Orderbook -> Create 2 Separate Orderbooks
+                        dataframes = []
+                        df_check_1 = df_temp.loc[df_temp.contract_id == contract_ids[0]]
+                        dataframes.append(df_check_1)
+                        df_check_2 = df_temp.loc[df_temp.contract_id == contract_ids[1]]
+                        df_check_2 = df_check_2.loc[df_check_2.as_of > f'{time - timedelta(**units)}']
+                        dataframes.append(df_check_2)
+
+                        # Quality Control For Temporary Dataframe
+                        temp_dataframe_list = []
+
+                        for nr, val in enumerate(dataframes):
+                            # Quality Control For Temporary Dataframe
+                            df_check = val.loc[~val.order_id.isin(orders_del)]
+
+                            if not df_check.empty:
+                                temp_dataframe_list.append(df_check)  # Add Filtered Df To List
+
+                        if len(temp_dataframe_list) == 1:
+                            order_book[f"{timestamp if timestamp and not from_timestamp else time}"] = \
+                                temp_dataframe_list[0]
+                        else:
+                            for nr, val in enumerate(temp_dataframe_list):
+                                if not nr:
+                                    order_book[
+                                        f"{(timestamp if timestamp and not from_timestamp else time) - timedelta(seconds=1)}"] = val
+                                else:
+                                    order_book[f"{timestamp if timestamp and not from_timestamp else time}"] = val
+
+                # Progress In Time Or Break Loop If Timestamp Exists
+                if timestamp and not from_timestamp:
+                    break
+
+                # Adjust Start Time To New Contract if necessary
+                if time == (delivery_start - timedelta(hours=1)):
+                    start_time = time
+
+                # Time Progression
+                time += timedelta(**units)
+
+            # General Quality Control
+            # Delete All Order ID Duplicates & Empty Timesteps
+            order_book_clean = {
+                key: value.sort_values(by=['as_of'], ascending=False).drop_duplicates(subset="order_id", keep="first",
+                                                                                      inplace=False) for (key, value) in
+                order_book.items()}
+            order_book_clean = {key: value for (key, value) in order_book_clean.items() if not value.empty}
+
+        else:
+            order_book_clean = df_bid_asks
+
+    orderbook_dict[key] = order_book_clean
+
+
 def _order_matching(order_side: str,
                     orderbook: pd.DataFrame,
                     timestamp: str,
@@ -251,6 +392,7 @@ def _order_matching(order_side: str,
                     exec_orders_list: dict[str, int],
                     trade_list: dict[int, dict[str, int]],
                     contract_time: int,
+					vwap: float = None,
                     order_execution: str = "NON") -> Decimal:
 	"""
 	Matches orders according to input parameters; adds trades made to trade_list and returns the remaining quantity.
@@ -275,6 +417,7 @@ def _order_matching(order_side: str,
 		exec_orders_list (dict): Dictionary of already matched order IDs
 		trade_list (list): List of executed trades
 		contract_time (int): contract time in minutes, either 60, 30 or 15
+		vwap (float): optional value to display current VWAP in the list of executed trades
 		order_execution (str): Type of order execution that should be simulated
 
 	Returns:
@@ -321,7 +464,8 @@ def _order_matching(order_side: str,
 				trade_list[len([*trade_list]) + 1] = {"Side": order_side,
 													  "Quantity": float(str(traded_quant)),
 													  "Price": row["price"],
-													  "Cash": float(str(cash)),
+													  "Cash": round(float(str(cash)), 2),
+													  "VWAP": round(vwap, 2),
 													  "Timestamp": timestamp}
 
 				if row["order_id"] in [*exec_orders_list]:
@@ -468,9 +612,9 @@ def _historic_data_transformation(files: list, exchange: str, filetype: str) ->
 							   "qty": 'quantity', "px": 'price', "pxqty": 'prc_x_qty'},
 					  inplace=True)
 
-			df.quantity = df.quantity / 100
+			df.quantity = df.quantity / 1000
 			df.price = df.price / 100
-			df.prc_x_qty = df.prc_x_qty / 10000
+			df.prc_x_qty = round(df.prc_x_qty / 100000, 2)
 
 		if filetype == "orders":
 			df.rename(columns={"asOf": "as_of", "bestAskPrice": "best_ask", "bestAskQuantity": "best_ask_qty",
@@ -498,7 +642,7 @@ def _historic_data_transformation(files: list, exchange: str, filetype: str) ->
 			asks = [i["sellOrdrList"]["ordrBookEntry"] if i["sellOrdrList"] else None for i in df.details.tolist()]
 			bids = [i["buyOrdrList"]["ordrBookEntry"] if i["buyOrdrList"] else None for i in df.details.tolist()]
 
-			conversion = lambda i: [{"order_id": v["ordrId"], "price": v["px"] / 100, "quantity": v["qty"] / 100,
+			conversion = lambda i: [{"order_id": v["ordrId"], "price": v["px"] / 100, "quantity": v["qty"] / 1000,
 									 "contract_id": contract_id, "delivery_area": delivery_area,
 									 "order_entry_time": v["ordrEntryTime"]} for v in i]
 
@@ -507,7 +651,7 @@ def _historic_data_transformation(files: list, exchange: str, filetype: str) ->
 
 			df["asks"] = asks
 			df["bids"] = bids
-
+			df.delta = [False if i else True for i in df.delta]
 			df.drop(columns=["_id", "details", "avwa", "bvwa"], inplace=True, errors="ignore")
 
 	else:
